# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for REANA-Client.

This file is imported by ``reana_client.__init__`` and parsed by
``setup.py``.
"""

from __future__ import absolute_import, print_function

__version__ = "0.95.0a4"
